@@ -31,7 +31,7 @@ from redssh import exceptions
 from redssh import enums
 from redssh import sftp
 from redssh import scp
-from redssh import tunnelling
+from redssh import tunneling
 from redssh import x11
 
 
@@ -46,7 +46,7 @@ class RedSSH(object):
     :type terminal: ``str``
     :param ssh_host_key_verification: Change the behaviour of remote host key verification. Can be set to one of the following values, ``strict``, ``warn``, ``auto_add`` or ``none``.
     :type ssh_host_key_verification: :class:`redssh.enums.SSHHostKeyVerify`
-    :param ssh_keepalive_interval: Enable or disable SSH keepalive packets, value is interval in seconds.
+    :param ssh_keepalive_interval: Enable or disable SSH keepalive packets, value is interval in seconds, ``0`` is off.
     :type ssh_keepalive_interval: ``float``
     :param set_flags: Not supported in ssh2-python 0.18.0
     :type set_flags: ``dict``
@@ -56,7 +56,7 @@ class RedSSH(object):
     def __init__(self,encoding='utf8',terminal='vt100',known_hosts=None
         ,ssh_host_key_verification=enums.SSHHostKeyVerify.warn,
         ssh_keepalive_interval=0.0,set_flags={},method_preferences={},
-        callbacks={}):
+        callbacks={},auto_terminate_tunnels=False):
         self.debug = False
         self._auto_select_timeout_enabled = False
         self._select_timeout = 0.005
@@ -77,6 +77,7 @@ class RedSSH(object):
         self.set_flags = set_flags
         self.method_preferences = method_preferences
         self.callbacks = callbacks
+        self.auto_terminate_tunnels = auto_terminate_tunnels
         self._ssh_keepalive_thread = None
         self._ssh_keepalive_event = None
         if known_hosts==None:
@@ -103,11 +104,13 @@ class RedSSH(object):
         while self.__check_for_attr__('channel')==False:
             time.sleep(timeout)
         while self._ssh_keepalive_event.is_set()==False and self.__check_for_attr__('channel')==True:
-            timeout = self._block(self.session.keepalive_send)
+            timeout = self._block(self.session.keepalive_send,_select_timeout=self._select_timeout)
             self._ssh_keepalive_event.wait(timeout=timeout)
 
 
-    def _block_select(self):
+    def _block_select(self,_select_timeout=None):
+        if _select_timeout==None:
+            _select_timeout = self._select_timeout
         with self._block_lock:
             block_direction = self.session.block_directions()
             if block_direction==0:
@@ -123,20 +126,27 @@ class RedSSH(object):
         else:
             wfds = []
 
-        select.select(rfds,wfds,[],self._select_timeout)
+        select.select(rfds,wfds,[],_select_timeout)
 
 
     def _block(self,func,*args,**kwargs):
         if self.__shutdown_all__.is_set()==False:
+            default_str = 'sdkljfhklsdjf'
+            _select_timeout = kwargs.get('_select_timeout',default_str)
+            if _select_timeout==default_str:
+                _select_timeout = None
+            else:
+                _select_timeout = float(_select_timeout)
+                del kwargs['_select_timeout']
             with self._block_lock:
                 out = func(*args,**kwargs)
             while out==libssh2.LIBSSH2_ERROR_EAGAIN:
-                self._block_select()
+                self._block_select(_select_timeout)
                 with self._block_lock:
                     out = func(*args,**kwargs)
             return(out)
 
-    def _block_write(self,func,data):
+    def _block_write(self,func,data,_select_timeout=None):
         data_len = len(data)
         total_written = 0
         while total_written<data_len:
@@ -145,10 +155,10 @@ class RedSSH(object):
                     (rc,bytes_written) = func(data[total_written:])
                 total_written+=bytes_written
                 if rc==libssh2.LIBSSH2_ERROR_EAGAIN:
-                    self._block_select()
+                    self._block_select(_select_timeout)
         return(total_written)
 
-    def _read_iter(self,func,block=False,max_read=-1):
+    def _read_iter(self,func,block=False,max_read=-1,_select_timeout=None):
         pos = 0
         remainder_len = 0
         remainder = b''
@@ -157,7 +167,7 @@ class RedSSH(object):
                 (size,data) = func()
             while size==libssh2.LIBSSH2_ERROR_EAGAIN or size>0:
                 if size==libssh2.LIBSSH2_ERROR_EAGAIN:
-                    self._block_select()
+                    self._block_select(_select_timeout)
                     if self.__shutdown_all__.is_set()==False:
                         with self._block_lock:
                             (size,data) = func()
@@ -173,7 +183,7 @@ class RedSSH(object):
                         else:
                             yield(data[pos:size])
                         pos = size
-                    self._block_select()
+                    self._block_select(_select_timeout)
                     with self._block_lock:
                         (size,data) = func()
                     pos = 0
@@ -211,9 +221,7 @@ class RedSSH(object):
                     if self._auth_attempt(self.session.userauth_password,username,password)==True:
                         return()
                 if 'keyboard-interactive' in auth_supported:
-                    auth_types_tried.append('keyboard-interactive') # not implemented in ssh2-python 0.18.0
-                    # bugged in ssh2-python's implementation for 1.9.0 of libssh2
-                    # but fixed in my fork. :)
+                    auth_types_tried.append('keyboard-interactive')
                     if self._auth_attempt(self.session.userauth_keyboardinteractive,username,password)==True:
                         return()
 
@@ -239,8 +247,7 @@ class RedSSH(object):
         Returns what value was settled on during session negotiation.
         '''
         if self.__check_for_attr__('session')==True:
-            if 'methods' in dir(self.session): # remove once my fork is merged.
-                return(self._block(self.session.methods,method))
+            return(self._block(self.session.methods,method))
 
     def setenv(self, varname, value):
         '''
@@ -255,6 +262,16 @@ class RedSSH(object):
         if self.__check_for_attr__('past_login'):
             self._block(self.channel.setenv,varname,value)
 
+    def check_closed(self,channel=None):
+        '''
+        Returns ``True`` or ``False`` when the main channel has recieved an ``EOF`` or an associated channel.
+        '''
+        if self.__check_for_attr__('session')==True and self.__check_for_attr__('channel')==True:
+            if channel==None:
+                return(self.eof()==True)
+            else:
+                return(self.eof()==True or self._block(channel.eof)==True)
+
     def check_host_key(self,hostname,port):
         if self.ssh_host_key_verification==enums.SSHHostKeyVerify.none:
             return(None)
@@ -337,17 +354,15 @@ class RedSSH(object):
             self.session = libssh2.Session()
             # self.session.publickey_init()
 
-            if 'flag' in dir(self.session): # remove once my fork is merged.
-                if not self.set_flags=={}:
-                    for flag in self.set_flags:
-                        self.session.flag(flag, self.set_flags[flag])
+            if not self.set_flags=={}:
+                for flag in self.set_flags:
+                    self.session.flag(flag, self.set_flags[flag])
 
-            if 'method_pref' in dir(self.session): # remove once my fork is merged.
-                if not self.method_preferences=={}:
-                    for pref in self.method_preferences:
-                        self.session.method_pref(pref, self.method_preferences[pref])
+            if not self.method_preferences=={}:
+                for pref in self.method_preferences:
+                    self.session.method_pref(pref, self.method_preferences[pref])
 
-            # if 'callback_set' in dir(self.session): # remove once my fork is merged.
+            # if 'callback_set' in dir(self.session):
                 # if not self.callbacks=={}:
                     # for cbtype in self.callbacks:
                         # self.session.callback_set(cbtype, self.callbacks[cbtype])
@@ -359,7 +374,7 @@ class RedSSH(object):
             self._auth(username,password,allow_agent,host_based,key_filepath,passphrase,look_for_keys)
 
             self.session.set_blocking(False)
-            if not self.ssh_keepalive_interval==0:
+            if self.ssh_keepalive_interval>0:
                 self.session.keepalive_config(True, self.ssh_keepalive_interval)
                 self._ssh_keepalive_thread = threading.Thread(target=self.ssh_keepalive)
                 self._ssh_keepalive_event = threading.Event()
@@ -368,7 +383,7 @@ class RedSSH(object):
             if self.request_pty==True:
                 self._block(self.channel.pty,self.terminal)
 
-            # if 'callback_set' in dir(self.session): # remove once my fork is merged.
+            # if 'callback_set' in dir(self.session):
                 # self._forward_x11()
 
             self._block(self.channel.shell)
@@ -403,6 +418,18 @@ class RedSSH(object):
                 return(self._block_write(self.channel.write,string))
         return(0)
 
+    def flush(self):
+        '''
+        Flush all data on the primary channel's stdin to the remote connection.
+        Only works if connected, otherwise returns ``0``.
+
+        :return: ``int`` - Amount of bytes sent to remote machine.
+        '''
+        if self.__check_for_attr__('past_login')==True:
+            if self.past_login==True:
+                return(self._block(self.channel.flush))
+        return(0)
+
     def last_error(self):
         '''
         Get the last error from the current session.
@@ -491,14 +518,14 @@ class RedSSH(object):
             wait_for_chan = threading.Event()
             thread_terminate = threading.Event()
 
-            class SubHander(tunnelling.LocalPortServerHandler):
+            class SubHander(tunneling.LocalPortServerHandler):
                 ssh_session = self
                 chain_host = remote_host
                 chain_port = remote_port
                 terminate = thread_terminate
                 wchan = wait_for_chan
 
-            tun_server = tunnelling.LocalPortServer((bind_addr,local_port),SubHander,self,thread_terminate,remote_host,remote_port,wait_for_chan,error_level)
+            tun_server = tunneling.LocalPortServer((bind_addr,local_port),SubHander,self,thread_terminate,remote_host,remote_port,wait_for_chan,error_level)
             tun_thread = threading.Thread(target=tun_server.serve_forever)
             tun_thread.daemon = True
             tun_thread.name = enums.TunnelType.local.value+':'+option_string
@@ -516,7 +543,7 @@ class RedSSH(object):
 
         Forwards a port to the remote machine via the local machine the same way the ``-R`` option does for the OpenSSH client.
 
-        :param local_port: The local port on the remote side to connect to.
+        :param local_port: The local port on the remote side for clients to connect to.
         :type local_port: ``int``
         :param remote_host: The remote host to connect to via the local machine.
         :type remote_host: ``str``
@@ -530,7 +557,7 @@ class RedSSH(object):
         if not option_string in self.tunnels[enums.TunnelType.remote.value]:
             wait_for_chan = threading.Event()
             thread_terminate = threading.Event()
-            tun_thread = threading.Thread(target=tunnelling.remote_tunnel_server,args=(self,remote_host,remote_port,bind_addr,local_port,thread_terminate,wait_for_chan,error_level))
+            tun_thread = threading.Thread(target=tunneling.remote_tunnel_server,args=(self,remote_host,remote_port,bind_addr,local_port,thread_terminate,wait_for_chan,error_level))
             tun_thread.daemon = True
             tun_thread.name = enums.TunnelType.remote.value+':'+option_string
             tun_thread.start()
@@ -559,14 +586,14 @@ class RedSSH(object):
             wait_for_chan = threading.Event()
             thread_terminate = threading.Event()
 
-            class SubHander(tunnelling.LocalPortServerHandler):
+            class SubHander(tunneling.LocalPortServerHandler):
                 ssh_session = self
                 chain_host = None
                 chain_port = None
                 terminate = thread_terminate
                 wchan = wait_for_chan
 
-            tun_server = tunnelling.LocalPortServer((bind_addr,local_port),SubHander,self,thread_terminate,None,None,wait_for_chan,error_level)
+            tun_server = tunneling.LocalPortServer((bind_addr,local_port),SubHander,self,thread_terminate,None,None,wait_for_chan,error_level)
             tun_thread = threading.Thread(target=tun_server.serve_forever)
             tun_thread.daemon = True
             tun_thread.name = enums.TunnelType.dynamic.value+':'+option_string
